/*
 * Component	: slist.c
 *
 * Description	: This file contains the structured list functions
 *			for the T/EC Adapters.
 *
 * Primary Author : Marc Foret
 *
 * Peer Reviewer  :
 *
 * (C) COPYRIGHT BIM, sa/nv 1994
 * Unpublished Work
 * All Rights Reserved
 * Licensed Material - BIM, sa/nv.
 *
 */

/*****************************************************************************/
/*                            I N T E R F A C E                              */
/*****************************************************************************/

/*** I M P O R T E D ***/

#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>

/* varargs.h is not supported on the AS/400 ilecx compiler */
#if defined(__STDC__) || defined(STDC_SRC_COMPATIBLE) || defined(__ILEC400__)
#	include <stdarg.h>
#else
#	include <varargs.h>
#endif

#if defined(NETWARE)

/*
** List traversals can take enough time that it makes the server
** unresponsive - joys of non-preemptive environment.
**
** Use a crude timing mechanism to yield the CPU once in a while.
*/
#if defined(ANCIENT_SDK)
#	include <process.h>
#else
#	include <nwthread.h>
#endif

#include <time.h>

#define QUANTUM 3

#define SETUP_TIMERS() \
    clock_t startTick = 0; \
    clock_t stopTick = 0

#define SET() startTick = clock()

#define YIELD() \
    stopTick = clock(); \
    if (stopTick - startTick > QUANTUM) { \
	ThreadSwitchWithDelay(); \
	startTick = clock(); \
    }

#endif /* defined(NETWARE) */

#include "tec_defines.h"
#include "slist.h"

CODE_ID("$Id: @(#)82	1.14  src/bim_server/tec_utils/slist.c, tec_utils, tec_3.8 5/12/00 15:27:22 $")

/* ------------------------------------------------------------------------- */
/*
 * Name: SL_InsLast
 *
 * Purpose: 	Add at the end of the list a new element.
 *
 * Input: 	handle:the handle of the list.
 *		elm: the element to add in the list.
 *
 * Ouput/Exceptions: MALLOC_ERROR: malloc error.
 * Returns: 	TRUE if success, FALSE otherwise.
 *
 */

int
SL_InsLast(SL_t *handle, void *elm)
{
	if (handle == NULL)
		return(FALSE);

	/*
	 * (re) allocate space
	 */
	if (handle->last_pos >= handle->nbr) {
		if (handle->nbr == 0)
			handle->list = (void *) malloc(sizeof(void *) * BLOCK_SIZE);
		else {
			handle->list = (void *) realloc(handle->list,
													  (sizeof(void *)
													   * (handle->nbr + BLOCK_SIZE)));
		}

		if (handle->list == NULL)
			return(FALSE);

		handle->nbr += BLOCK_SIZE;
	}
	handle->list[handle->last_pos++] = elm;

	return(TRUE);
}

/* ------------------------------------------------------------------------- */

/*
 * Name: SL_Find
 *
 * Purpose: 	Find a element in the list.
 *
 * Input: 	handle:the handle of the list.
 *		cmp_data: the data to compare in the list.
 *		cmp_func:	the comparaison func to apply.
 *
 * Ouput/Exceptions: MALLOC_ERROR: malloc error.
 * Returns: 	elem if cmp_data success, NULL otherwise.
 *
 */

void *
SL_Find(SL_t *handle, void *cmp_data, SL_compare_fn cmp_func)
{
	int	pos = SL_FindPos(handle, cmp_data, cmp_func);

	/*
	 * Don't put "SL_FindPos(handle, cmp_data, cmp_func)" directly into
	 * the SL_GetPos call because SL_GetPos is a macro and SL_FindPos would
	 * get called twice.
	 */
	return(SL_GetPos(handle, pos));
}

/* ------------------------------------------------------------------------- */

/*
 * Name: SL_FindPos
 *
 * Purpose: 	Find the position of a element in the list.
 *
 * Input: 	handle:the handle of the list.
 *		cmp_data: the data to compare in the list.
 *		cmp_func:	the comparaison func to apply.
 *
 * Ouput/Exceptions: MALLOC_ERROR: malloc error.
 * Returns: 	pos if cmp_data success, -1 otherwise.
 *
 */

int
SL_FindPos(SL_t *handle, void *cmp_data, SL_compare_fn cmp_func)
{
#if defined(NETWARE)
    SETUP_TIMERS();
#endif /* defined(NETWARE) */

	register int	i;

	/* SL_errno = INVALID_STRUCTUR */
	if ((handle == NULL) || (handle->nbr == 0) || (cmp_func == NULL))
		return(-1);

#if defined(NETWARE)
    SET();
#endif /* defined(NETWARE) */

	for (i = 0; i < handle->last_pos; i++) {
		if (cmp_func(handle->list[i], cmp_data) == TRUE)
			return(i);

#if defined(NETWARE)
	YIELD();
#endif /* defined(NETWARE) */

	}
	return(-1);
}

/* ------------------------------------------------------------------------- */

/*
 * Name: SL_FindReplace
 *
 * Purpose: 	Find and Replace an element in the list.
 *
 * Input: 	handle:		the handle of the list.
 *		replace_data: 	the data to compare/replace in the list.
 *		cmp_func:	the comparison func to apply.
 *		free_func: 	a free function of the data.
 *
 * Ouput/Exceptions: MALLOC_ERROR: malloc error.
 * Returns: 	TRUE if success, FALSE otherwise.
 *
 */

int
SL_FindReplace(SL_t *handle, void *replace_data, SL_compare_fn cmp_func,
					SL_free_fn free_func)
{
	register int	i;

#if defined(NETWARE)
    SETUP_TIMERS();
#endif /* defined(NETWARE) */

	if ((handle == NULL) || (handle->nbr == 0) || (cmp_func == NULL))
		return(FALSE);

#if defined(NETWARE)
    SET();
#endif /* defined(NETWARE) */

	for (i = 0; i < handle->last_pos; i++) {
		if (cmp_func(handle->list[i], replace_data) == TRUE) {
			free_func(handle->list[i]);
			handle->list[i] = replace_data;
			return(TRUE);
		}

#if defined(NETWARE)
	YIELD();
#endif /* defined(NETWARE) */

	}
	return(FALSE);
}

/* ------------------------------------------------------------------------- */

/*
 * Name: SL_DelAll
 *
 * Purpose: 	Destroys a list:.
 *
 * Input: 	handle:the handle of the list.
 *		free_func: a free function of the data.
 *
 * Ouput/Exceptions:
 *
 * Returns: 	/
 *
 */

void
SL_DelAll(SL_t *handle, SL_free_fn free_func)
{
	int i;

#if defined(NETWARE)
    SETUP_TIMERS();
#endif /* defined(NETWARE) */

	if (handle == NULL)
		return;

	if (handle->list != NULL) {

#if defined(NETWARE)
	SET();
#endif /* defined(NETWARE0 */

		for (i = 0; i < handle->last_pos; i++) {
			free_func(handle->list[i]);

#if defined(NETWARE)
	  YIELD();
#endif /* defined(NETWARE) */

		}
		free(handle->list);
	}
	free(handle);

	return;
}

int
SL_Delete(SL_t *handle, void *data, SL_free_fn func)
{
	register int	i;

#if defined(NETWARE)
    SETUP_TIMERS();
#endif /* defined(NETWARE) */

	if ((handle == NULL) || (handle->nbr == 0))
		return(-1);

#if defined(NETWARE)
	SET();
#endif /* defined(NETWARE0 */

	for (i = 0; (i < handle->last_pos); i++) {
		if (handle->list[i] == data) {
			if (func != NULL)
				func(handle->list[i]);

			memmove(&(handle->list[i]), &(handle->list[i + 1]),
                 (sizeof(void *) * (handle->last_pos - (i + 1))));
			handle->last_pos--;

			return(i);
		}

#if defined(NETWARE)
	  YIELD();
#endif /* defined(NETWARE) */

	}
	return(-1);
}

/* ------------------------------------------------------------------------- */

/*
 * Name: SL_Walk
 *
 * Purpose: 	For each element in the list, call (func) with argument
 *		data: element in the list and the position in the list.
 *
 * Input: 	handle:the handle of the list.
 *		func: the func to call for each element in the list.
 *
 * Ouput/Exceptions: if the func return 0 stops the walk with no error.
 *		if the func return -1 stops the walk with error.
 *
 *
 * Returns: 	FALSE if error, TRUE otherwise.
 *
 */

int
SL_Walk(SL_t *handle, SL_walk_fn func)
{
#if defined(NETWARE)
    SETUP_TIMERS();
#endif /* defined(NETWARE) */

	register int	i, ret;

	if ((handle == NULL) || (handle->nbr == 0) || (func == NULL))
		return(FALSE);

#if defined(NETWARE)
    SET();
#endif /* defined(NETWARE) */

	for (i = 0; i < handle->last_pos; i++) {
		ret = func(handle->list[i], (i + 1));

		if (ret == 0)
			return(TRUE);
		else if (ret == -1)
			return(FALSE);

#if defined(NETWARE)
	YIELD();
#endif /* defined(NETWARE) */

	}
	return(FALSE);
}


/*
 * Name: SL_Walk2
 *
 * Purpose: 	For each element in the list, call (func) with argument
 *		data: element in the list and the position in the list.
 *
 * Input: 	handle:the handle of the list.
 *		func: the func to call for each element in the list.
 *
 * Ouput/Exceptions: if the func return 0 stops the walk with no error.
 *		if the func return -1 stops the walk with error.
 *
 *
 * Returns: 	FALSE if error, TRUE otherwise.
 *
 */

int
SL_Walk2(SL_t *handle, SL_walk2_fn func, void *arg)
{
	register int i, ret;

#if defined(NETWARE)
    SETUP_TIMERS();
#endif /* defined(NETWARE) */

	if ((handle == NULL) || (handle->nbr == 0) || (func == NULL))
		return(FALSE);

#if defined(NETWARE)
    SET();
#endif /* defined(NETWARE) */

	for (i = 0; i < handle->last_pos; i++) {
		ret = func(handle->list[i], arg);

		if (ret == 0)
			return(TRUE);
		else if (ret == -1)
			return(FALSE);

#if defined(NETWARE)
	YIELD();
#endif /* defined(NETWARE) */

	}
	return(FALSE);
}


/* ------------------------------------------------------------------------- */

/*
 * Name: SL_WalkVa
 *
 * Purpose: 	Return a pointer to the n-th element of the list.
 *
 * Input: 	handle:the handle of the list.
 *		func: the func to call for each element in the list.
 *		...: Addition args to pass with the func. in the list.
 *
 * Ouput/Exceptions: if the func return 0 stops the walk with no error.
 *		if the func return -1 stops the walk with error.
 *
 *
 * Returns: 	FALSE if error, TRUE otherwise.
 *
 */

#if !defined(__ILEC400__)     /* This function is currently not supported */

#if defined(__STDC__) || defined(STDC_SRC_COMPATIBLE)

int
SL_WalkVa(SL_t *handle, int (*func)(), ...)

#else

int
SL_WalkVa(SL_t *handle, SL_walkva_fn func, va_dcl)

#endif
{
	register int		i, ret;
	va_list	args;

#if defined(NETWARE)
    SETUP_TIMERS();
#endif /* defined(NETWARE) */

#if defined(__STDC__) || defined(STDC_SRC_COMPATIBLE)
	va_start(args, func);
#else
	va_start(args);
#endif

	if ((handle == NULL) || (handle->nbr == 0) || (func == NULL)) {
		va_end(args);
		return(FALSE);
	}

#if defined(NETWARE)
    SET();
#endif /* defined(NETWARE) */

	for (i = 0; i < handle->last_pos; i++) {
		if ((ret = func(handle->list[i], args)) == 0) {
			va_end(args);
			return(TRUE);
		}
		else if (ret == -1) {
			va_end(args);
			return(FALSE);
		}

#if defined(NETWARE)
	YIELD();
#endif /* defined(NETWARE) */

	}
   va_end(args);

	return(FALSE);
}

#endif    /* end of:  if !defined(__ILEC400__)  */

/* ------------------------------------------------------------------------- */

SA_t *
SA_Create(int elt_size)
{
	SA_t	*list = (SA_t *) calloc(1, sizeof(SA_t));

	list->elt_size = elt_size;

	return(list);
}

/*
 * Name: SA_InsLast
 *
 * Purpose: 	Add at the end of the list a new element.
 *
 * Input: 	handle:the handle of the list.
 *		elm: the element to add in the list.
 *
 * Ouput/Exceptions: MALLOC_ERROR: malloc error.
 * Returns: 	TRUE if success, FALSE otherwise.
 *
 */

void *
SA_InsLast(SA_t *handle, void *elm)
{
	void	*dest_slot;

	if (handle == NULL)
		return(NULL);

	/*
	 * (re) allocate space
	 */
	if (handle->last_pos >= handle->nbr) {
		if (handle->nbr == 0)
			handle->list = calloc(BLOCK_SIZE, handle->elt_size);
		else {
			handle->list = realloc(handle->list, (handle->elt_size
															  * (handle->nbr + BLOCK_SIZE)));
		}

		if (handle->list == NULL)
			return(NULL);

		handle->nbr += BLOCK_SIZE;
	}
	dest_slot = (void *) ((char *) handle->list + (handle->last_pos++ * handle->elt_size));
	memcpy(dest_slot, elm, handle->elt_size);;

	return(dest_slot);
}

void
SA_Delete(SA_t *handle, void *el_ptr)
{
	void	*end_ptr;

	if ((handle == NULL) || (handle->nbr == 0))
		return;

	end_ptr = (void *) ((char *) handle->list + (handle->last_pos * handle->elt_size));

	if ((el_ptr < handle->list) || (el_ptr > end_ptr))
		return;

	memcpy(el_ptr, (void *) ((char *) el_ptr + handle->elt_size),
			 ((size_t) end_ptr - (size_t) el_ptr - handle->elt_size));
	handle->last_pos--;
}

/*
 * Name: SA_Walk2
 *
 * Purpose: 	For each element in the list, call (func) with argument
 *		data: element in the list and the position in the list.
 *
 * Input: 	handle:the handle of the list.
 *		func: the func to call for each element in the list.
 *
 * Ouput/Exceptions: if the func return 0 stops the walk with no error.
 *		if the func return -1 stops the walk with error.
 *
 *
 * Returns: 	FALSE if error, TRUE otherwise.
 *
 */

int
SA_Walk2(SA_t *handle, SA_walk2_fn func, void *arg)
{
	register int i, ret;

	if ((handle == NULL) || (handle->nbr == 0) || (func == NULL))
		return(FALSE);

	for (i = 0; i < handle->last_pos; i++) {
		ret = func((void *) ((char *)handle->list + (handle->elt_size * i)), arg);

		if (ret == 0)
			return(TRUE);
		else if (ret == -1)
			return(FALSE);
	}
	return(FALSE);
}
