/*
 * Component Name: EIF sockets library
 *
 * $Date: 9/3/02 09:35:56 $
 *
 * $Source: src/eif/sockeif.c, tec_eif, tec_3.8 $
 *
 * $Revision: 1.14.1.23 $
 *
 * Description: This library implements a few IPC functions for the TEC/EIF
 *  so that we don't have to ship the entire ADE.  This file must NOT be
 *  linked with ADE libraries, as the data structures are not compatible.
 *  These functions can be used to talk to ADE-type entities, as the protocol
 *  across the wire is consistent.
 *
 * (C) Copyright 1990-1994 Tivoli Systems, Inc.
 * Unpublished Work
 * All Rights Reserved
 * Licensed Material - Property of Tivoli Systems, Inc.
 *
 * =======================
 * NOTE: Set tab stops = 3
 * =======================
 */

#include <sys/types.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#if defined(AIX) || defined(WIN32) || defined(HP_UX) || defined(LINUX)
#include <fcntl.h>
#endif

#ifdef RELIANT_UNIX
/* Pull in TCP_NODELAY definition */
#define XTI_SUPPORT
#	include <netinet/tcp.h>
#endif

#ifdef AIX
#       include <sys/select.h>
#endif

#ifdef OS2
#	define TCP_NODELAY     0x01    /* from netinet/tcp.h */
#	include <signal.h>
#	include <nerrno.h>
#	include <utils.h>
#	define EINTR SOCEINTR
#else
#	include <errno.h>
#	ifndef PC
#		include <unistd.h>
#	elif !defined(NW4)
#		include "cpl/unistd.h"
#	endif
#endif

#ifdef NETWARE
#	include <io.h>  /* for close() */
#endif

#include "tec_rpc.h"

/*#if !defined(OS2) && !defined(WIN32)*/
#ifndef PC
#	include <netinet/tcp.h>
#	include <sys/ioctl.h>
#endif

#include "nontme_eif.h"
#ifdef OS400
   #include "qymacnv.h"
#endif

#define TEC_SERVER_NAME		"tec_server"
#define TEC_SERVER_NUMBER	100033057
#define TEC_VERSION_NUMBER	1

/*
** IY25233.1 - port 136491
*/

extern int EIPC_VERIFY_PORTMAPPER;


static int
set_ipc_flags(int fd)
{
#if !defined(NETWARE)
	struct protoent	*proto_info;
	int					true_flag;
	int					rc;

	/*
	 * Need prototype number for TCP before setting TCP_NODELAY option.
	 * Then set the flag.
	 * Finally, get the flag to confirm it was changed.
	 */
	if ((proto_info = getprotobyname("tcp")) != NULL) {
		true_flag = 1;

#		ifdef WIN32
			rc = setsockopt((SOCKET) fd, proto_info->p_proto, TCP_NODELAY,
								 (char FAR *) &true_flag, sizeof(true_flag));
#		else
			rc = setsockopt(fd, proto_info->p_proto, TCP_NODELAY,
								 (void *) &true_flag, (int) sizeof(true_flag));
#		endif
	}
	else {
		rc = -1;
	}

	return(rc);
#else  /* defined(NETWARE) */
    return -1;
#endif /* !defined(NETWARE) */
}

/*
 * Description:  create ipc client handle
 *
 * Returns: handle on success, NULL on failure
 *
 * Notes/Dependencies:
 */

eipc_handle_p_t
eipc_create_remote_client(unsigned addr, unsigned short port, eipc_error_t *error)
{
	eipc_handle_p_t     handle;
	struct sockaddr_in server;
	struct rpcent	*rpcinfo;
	u_long 			prognum;

#ifdef OS400
	handle = (eipc_handle_p_t)malloc(sizeof(eipc_handle_t));
	if ( NULL == handle ) {
		*error = E_NO_MEM;
		return 0;
	}
	handle->fd = socket(AF_INET, SOCK_STREAM, 0);
#else
	int ret;

	handle = (eipc_handle_p_t) malloc(sizeof(eipc_handle_t));
	if ( NULL == handle ) {
		*error = E_NO_MEM;
		return 0;
	}

	/*
	 * Don't let SIGALRM break this
	 */
	errno = EINTR;
	handle->fd = -1;
	while ((errno == EINTR) && (handle->fd == -1)) {
	   errno = 0;
	   handle->fd = socket(AF_INET, SOCK_STREAM, 0);
	}
#endif  /* ifdef OS400 */

	if (handle->fd < 0) {
		free(handle);
		*error = E_SCALL;
		return 0;
	}

	set_ipc_flags(handle->fd);

	memset((char*)&server,0,sizeof(server));
	server.sin_family = AF_INET;
	server.sin_port = port;
	server.sin_addr.s_addr = addr;

#ifdef OS400
	if (connect(handle->fd, (struct sockaddr*)&server, sizeof(server)) < 0) {
#else
	/*
	 * Don't let SIGALRM break this
	 */
	errno = EINTR;
	ret = -1;

	while ((errno == EINTR) && (ret == -1)) {
	   errno = 0;
		/*
		 ** IY25233.1
		 **
		 ** Some customers don't want to wait if the server being contacted is down
		 ** they want to simply cache until we can show the server has a reasonable
		 ** chance of being on line.  How to do this?  By infering that the receiving
		 ** tec server is up based on portmapper having a registered port for the
		 ** tec server, TEC_SERVER_NUMBER, we can have a reasonable expectation that
		 ** the receiving server is on line.  The implication of this is that the
		 ** target server has portmapper running and that "tec_rule_use_portmapper="
		 ** is not set or is not set to anything other than "YES" in the .tec_config.
		 **
		 ** Since portmapper is unavailable on NT systems, it is further implied that
		 ** the "EIPC_VERIFY_PORTMAPPER" flag won't be enabled on other endpoints, MN etc.
		 ** when they are communicating with an NT EventServer.
		 **
		 */
		if (EIPC_VERIFY_PORTMAPPER)
		{

			/*
			 * Get the program number for the call to tec_pmap_getport.
			 */
#ifdef PC
			prognum = TEC_SERVER_NUMBER;
#else
			rpcinfo = (struct rpcent *)NULL;
			errno = EINTR;
			while ((errno == EINTR)&&(rpcinfo == (struct rpcent *)NULL)) {
				errno = 0;
				rpcinfo = getrpcbyname(TEC_SERVER_NAME);
			}
			prognum = (rpcinfo == NULL) ? TEC_SERVER_NUMBER : rpcinfo->r_number;
#endif

			/*
			 * IY25233 - Set the port to 0 before the tec_pmap_getport call since it
			 * returns a port. This call is only used in order to see if the system
			 * that we're trying to send to is up. Without this call, if the system
			 * is down, the connect call (below) will hang the system for 75 seconds
			 * every time it's called (for each event).
			 */
			server.sin_port = 0;
#if !defined(WIN32) && !defined(OS2) && !defined(NW4)
			ret = tec_pmap_getport(&server, prognum, TEC_VERSION_NUMBER, IPPROTO_TCP);
#else
			ret = pmap_getport(&server, prognum, TEC_VERSION_NUMBER, IPPROTO_TCP);
#endif

			/*
			 * If tec_pmap_getport returns 0, we assume the system is down.
			 */
			if ( ret <= 0 )
				return(0);

		   server.sin_port = port;

		} /* if EIPC_VERIFY_PORTMAPPER */

		ret = connect(handle->fd, (struct sockaddr*)&server, sizeof(server));
	}

	if (ret < 0) {
#endif
#		if defined(PC) && !defined(NETWARE)
#			if !defined(OS2)
	                        *error = (WSAGetLastError() == WSAECONNREFUSED)
				  ? E_IPC_BROKEN : E_SCALL;
				closesocket(handle->fd);
				free(handle);
#			else
				*error = (errno == ECONNREFUSED) ? E_IPC_BROKEN : E_SCALL;
				soclose(handle->fd);
				free(handle);
#			endif
#		else
			*error = (errno == ECONNREFUSED) ? E_IPC_BROKEN : E_SCALL;
			close(handle->fd);
			free(handle);
#		endif

		return 0;
	}
	return handle;
}

/*
 * Description:  create ipc client handle
 *
 * Returns: handle on success, NULL on failure
 *
 * Notes/Dependencies:
 */

eipc_handle_p_t
gw_eipc_create_remote_client(unsigned addr, unsigned short port, eipc_error_t *error)
{
	eipc_handle_p_t     handle;
	struct sockaddr_in server;
	struct rpcent	*rpcinfo;
	u_long 			prognum;

#ifdef OS400
	handle = (eipc_handle_p_t)malloc(sizeof(eipc_handle_t));
	if ( NULL == handle ) {
		*error = E_NO_MEM;
		return 0;
	}
	handle->fd = socket(AF_INET, SOCK_STREAM, 0);
#else
	int ret;

	handle = (eipc_handle_p_t) malloc(sizeof(eipc_handle_t));
	if ( NULL == handle ) {
		*error = E_NO_MEM;
		return 0;
	}

	/*
	 * Don't let SIGALRM break this
	 */
	errno = EINTR;
	handle->fd = -1;
	while ((errno == EINTR) && (handle->fd == -1)) {
	   errno = 0;
	   handle->fd = socket(AF_INET, SOCK_STREAM, 0);
	}
#endif  /* ifdef OS400 */

	if (handle->fd < 0) {
		free(handle);
		*error = E_SCALL;
		return 0;
	}

	set_ipc_flags(handle->fd);

	memset((char*)&server,0,sizeof(server));
	server.sin_family = AF_INET;
	server.sin_port = port;
	server.sin_addr.s_addr = addr;

	if (connect(handle->fd, (struct sockaddr*)&server, sizeof(server)) < 0) {
#		if defined(PC) && !defined(NETWARE)
#			if !defined(OS2)
	                        *error = (WSAGetLastError() == WSAECONNREFUSED)
				  ? E_IPC_BROKEN : E_SCALL;
				closesocket(handle->fd);
				free(handle);
#			else
				*error = (errno == ECONNREFUSED) ? E_IPC_BROKEN : E_SCALL;
				soclose(handle->fd);
				free(handle);
#			endif
#		else
				/*			*error = (errno == ECONNREFUSED) ? E_IPC_BROKEN : E_SCALL;*/
				*error = errno;
			close(handle->fd);
			free(handle);
#		endif

		return 0;
	}
	return handle;
}

/*
 * Description: graceful shutdown of communications
 *
 * Returns: void
 *
 * Notes/Dependencies:
 */
void
eipc_shutdown(eipc_handle_p_t ipch)
{
    if (ipch->fd >= 0) {
        shutdown( ipch->fd, 2);
    }
}

/*
 * Description: close and free handle
 *
 * Returns: void
 *
 * Notes/Dependencies: input pointer is freed before return
 */
void
eipc_destroy_handle(eipc_handle_p_t ipch)
{
    if (ipch->fd >= 0) {
#if !defined(PC) || defined(NETWARE)
        close(ipch->fd);
#else
#	ifndef OS2
		closesocket(ipch->fd);
#	else
		soclose(ipch->fd);
#	endif
#endif
    }
    free(ipch);
}

/*
 * Description: socket-level send function
 *
 * Returns: 0 for success, -1 for failure
 *
 * Notes/Dependencies:
 */
static int
do_send(int fd, char *buf, int len, eipc_error_t *error)
{
    int rv;
retry:
    rv = send(fd, buf, len, 0);
    if (rv < 0) {
        if (errno == EINTR) {
            goto retry;
        }
        *error = E_SCALL;
        return -1;
    }
    if (rv < len) {
        len -= rv;
        buf += rv;
        goto retry;
    }
    return 0;
}

/*
 * Description:  Send a message
 *
 * Returns: 0 for success, -1 for failure
 *
 * Notes/Dependencies:  Compatible message format with libtmf's and
 *  libwo's ipc_send.
 */

int
eipc_send(eipc_handle_p_t handle, char *data, long len, long type, long id,
                                                          eipc_error_t *error)
{
    eipc_msg_hdr_t hdr;
    int           dlen;
    int           slen;
    char*         data_ascii;
    int           convert_msg = FALSE;

    /*
    ** Put this check in to avoid sending headers without messages.
    ** T/EC 3.6 server seems to have difficulty dealing with it
    ** when in connection-oriented (CO) mode.
    */
    if (len < 1) {
        return 0;
    }

    memcpy(hdr.token, TOKEN, 8);
    hdr.msg_id = htonl(id);
    hdr.msg_from = 0;
    hdr.msg_to = 0;
    hdr.msg_type = htonl(type);
    hdr.ipc_msg_type = 0;
    hdr.msg_len = htonl(len);

#ifdef OS400
    /* Convert the Event Message from EBCDIC to ASCII.
     * This is *ONLY* done if the message is going to a pre 3.7 server,
     * and the "Pre37Server=YES" line is in the configuration file.
     * iconCd variable is defined in header qymacnv.h
     */
    if (data[0] != '\x03')
    {
       convert_msg = TRUE;
       data_ascii = strdup(data);
	   if ( data_ascii == NULL) {
		   return -1;
	   }

       QymaCodeConversion(iconCd, data_ascii, data_ascii, &len);
     }
#endif

    if(len < HDR_DATA_LEN) {
        dlen = len;
    } else {
        dlen = HDR_DATA_LEN;
    }
    if((hdr.hdr_data_len = htonl(dlen)) != 0) {
#ifdef OS400
        if (convert_msg)
            memcpy(hdr.hdr_data, data_ascii, dlen);
        else
            memcpy(hdr.hdr_data, data, dlen);
#else
        memcpy(hdr.hdr_data, data, dlen);
#endif
    }
    slen = sizeof(eipc_msg_hdr_t) - HDR_DATA_LEN + dlen;

    /* send the header (and data if it's short) */
    if (do_send(handle->fd, (char*)&hdr, slen, error) != 0) {
        return -1;
    }

    /* send the rest of the message body (if it was longer than
     * HDR_DATA_LEN
     */
    if (len - dlen > 0) {
#ifdef OS400
        if (convert_msg) {
            if (do_send(handle->fd, data_ascii + dlen, len - dlen, error) != 0) {
                return -1;
            }
        }
        else {
            if (do_send(handle->fd, data + dlen, len - dlen, error) != 0) {
                return -1;
            }
        }
#else
        if (do_send(handle->fd, data + dlen, len - dlen, error) != 0) {
            return -1;
        }
#endif
    }
#ifdef OS400
    if (data_ascii != NULL)
        free(data_ascii);
#endif

    return 0;
}

/*
 * Description: socket-level receive function
 *
 * Returns: 0 for success, -1 for failure
 *
 * Notes/Dependencies:
 */
static int
do_recv(int fd, char * buf, int len, eipc_error_t *error)
{
#if defined(AIX) || defined(WIN32) || defined(HP_UX) || defined(LINUX)
    int brecv = 0;

    while (brecv < len) {
      brecv = recv(fd, buf + brecv, len - brecv, 0);

      /* if no bytes could be read, it means the writer is not putting */
      /* anything there, which is therefore a problem in the reception */
      /* 147336 - dnastaci */
      if (brecv == 0) {
	*error = -1;   /* there is really no standard error code to put here */
        return -1;
      }

      if (brecv < 0) {
        if (errno == EINTR) {
	  continue;
        }
        /**error = E_SCALL;*/
        *error = errno;
        return -1;
      }
    }
#endif
    return 0;
}


/*
 * Description:  Receive a message
 *
 * Returns: the message for success, NULL for failure
 *
 * Notes/Dependencies:  Compatible message format with libtmf's and
 *  libwo's ipc_send.
 */

char *
eipc_recv(eipc_handle_p_t handle, eipc_error_t *error)
{
    char *msg = NULL;
#if defined(AIX) || defined(WIN32) || defined(HP_UX) || defined(LINUX)
    int msg_len;
    eipc_msg_hdr_t hdr;
    int hdr_len = sizeof(eipc_msg_hdr_t) - sizeof(char) * HDR_DATA_LEN;

    /* if an error occurr while reading the header, return */
    if (do_recv(handle->fd, (char*)&hdr, hdr_len, error) == -1) {
	return NULL;
    }

    /* retrieve the number of bytes to read */
    msg_len = htonl(hdr.msg_len);
    if (msg_len > 0) {
	msg = (char*) malloc (msg_len + 1);

	/* if an error occurr while reading the message, return */
	if (do_recv(handle->fd, msg, msg_len, error) == -1) {
	    free(msg);
	    return NULL;
	}

	msg[msg_len] = '\0';
    }

#endif
    return msg;
}

/*
 * Description:  Check if there is a message available for reading
 *
 * Returns: 1 for success - data is ready to be read, 0 for failure
 *
 * Notes/Dependencies:  Compatible message format with libtmf's and
 *  libwo's ipc_send.
 */

int
eipc_select(eipc_handle_p_t handle, eipc_error_t *error)
{
  int rc = 0;
#if defined(AIX) || defined(WIN32) || defined(HP_UX) || defined(LINUX)
  fd_set read_fds;
  /* do not block */
  struct timeval timeout = {0, 0};

  FD_ZERO(&read_fds);
  FD_SET(handle->fd,&read_fds);

  rc = select(FD_SETSIZE, &read_fds, NULL, NULL, &timeout);

  if (rc > 0) {
    rc = 1;
  }
  else {
    *error = 0;
    if (rc == -1) {
      *error = errno;
    }
    rc = 0;
  }
#endif

  return rc;
}

                                                                  /*S-147204 JHC */
/*
 * Description:  Check if there is a message available for reading
 *
 * Returns: 1 for success - data is ready to be read, 0 for failure
 *
 * Notes/Dependencies:  Compatible message format with libtmf's and
 *  libwo's ipc_send.
 */

int
eipc_select_wait(eipc_handle_p_t handle, eipc_error_t *error, int s, int u)
{
  int rc = 0;
#if defined(AIX) || defined(WIN32) || defined(HP_UX) || defined(LINUX)
  fd_set read_fds;
  /* do not block */
  struct timeval timeout = {s, u};

  FD_ZERO(&read_fds);
  FD_SET(handle->fd,&read_fds);

  rc = select(FD_SETSIZE, &read_fds, NULL, NULL, &timeout);

  if (rc > 0) {
    rc = 1;
  }
  else {
    *error = 0;
    if (rc == -1) {
      *error = errno;
    }
    rc = 0;
  }
#endif

  return rc;
}
                                                               /*E-147204 JHC */
/*
 * Description:  Set the reading to NON_BLOCKING
 *
 * Returns: 1 for success - data is going to be read async, 0 for failure
 *
 * Notes/Dependencies:  Compatible message format with libtmf's and
 *  libwo's ipc_send.
 */
/* int eipc_async(eipc_handle_p_t handle, eipc_error_t *error)
{
#if defined(AIX) || defined(HP_UX) || defined(LINUX)
    if (fcntl(handle->fd, F_SETFL, O_NONBLOCK) < 0) {
	*error = errno;
	return 0;
    }
#endif
    return 1;
}
*/

int eipc_async(eipc_handle_p_t handle, eipc_error_t *error)
{
	 int rc=0;

#if defined(LINUX)
   rc= fcntl(handle->fd, F_SETFL, FNDELAY);
   *error = errno;
#endif

#if defined(HP_UX)
   rc = fcntl(handle->fd, F_SETFL, O_NONBLOCK);
   *error = errno;
#endif

#if defined(WIN32)
	{
	u_long active = 1;
	int timeout = 1000 * 60 ;
	if ((rc = ioctlsocket(handle->fd, FIONBIO, &active)) == SOCKET_ERROR)
	{
		/* ERR "ioctlsocket" END; */
	}
	}
	*error = WSAGetLastError();

#endif

#if defined(AIX)
	{
   int   no_block = 1;
   rc = ioctl(handle->fd, FIONBIO, (char *)&no_block);
   *error = errno;
	}
#endif

return rc;
}
