/*
 * Component Name:  EIF
 *
 * Dependencies:
 *
 * Description:
 *    Public header file for EIF interface.  This file should contain all
 *    definitions and prototypes needed to use the EIF library.
 *    Conversely, it should not contain any thing that is not intended to
 *    be public.
 *
 * (C) COPYRIGHT TIVOLI Systems, Inc. 1999
 * Unpublished Work
 * All Rights Reserved
 * Licensed Material - Property of TIVOLI Systems, Inc.
 */

#ifndef TEC_EIF_H
#define TEC_EIF_H

#ifndef PC
#	include <sys/time.h>
#endif

#include "tec_defines.h"

#define EVENT_MAX_SIZE        "EventMaxSize"
#define FILTER                "Filter:"
#define FILTER_CACHE          "FilterCache:"
#define TECAD_FILTERMODE      "FilterMode"
#define FILE_LEN              256
#define PREFILTER             "PreFilter:"
#define TECAD_PFLTMODE        "PreFilterMode"
#define TECAD_EVENT_TO_CATCH  "NumEventsToCatchUp"
#define VERSION_3_SERVER_ENCODING  "Pre37ServerEncoding"
#define VERSION_3_SERVER_COMPATIBILITY  "Pre37Server"
#define WORKBUF_SIZE          4096
#define MAX_PACKET_SIZE    "MaxPacketSize"

/*
 * All TEC events must end with TECAD_EVENT_DELIMITER
 */
#if defined(OS400)
   #define TECAD_EVENT_DELIMITER			"END\x025\001"
#else
   #define TECAD_EVENT_DELIMITER			"END\n\001"
#endif
#define TECAD_EVENT_DELIMITER_LEN	5
#define TECAD_EVENT_END_CHAR			'\001'

/*
 * Types of event filters.
 *
 * FILTER_EVENT_TYPE         Filters events into/from being sent to the server
 * FILTER_EVENT_CACHE_TYPE   Filters events into/from being buffered if server
 *                           is unavailable.
 */
typedef enum {
    FILTER_EVENT_TYPE,
    FILTER_EVENT_CACHE_TYPE
} FilterType;

/*
 * Types of connections to the TEC Server.
 *
 * connection_less      A new connection is opened and closed for each event
 * connection_oriented  One open connection is maintained for sending all events
 * use_default          Use the default value specified in the EIF *.conf file
 *                      with the "ConnectionMode" setting.
 */
typedef enum {
	connection_less,
	connection_oriented,
	use_default
} tec_delivery;

typedef void *tec_handle_t;

#ifdef OS400
   extern int startup_msgs;
#endif

/*
 * Global variables.
 */
EXTERN_VAR(int tec_errno);
EXTERN_VAR(int TECAD_no_connection);

/*
 * Global functions.
 */

/*
 * To initialize configuration parameters from the given configuration file
 * (typically named with a .conf extension).  tec_method_agent_init is a
 * variation that is used by processes that are CORBA method servers.
 */
EXTERN_DECL(int tec_agent_init, (char *));
EXTERN_DECL(int tec_method_agent_init, (char *));

/*
 * Retrieves the value of a given variable that was read from a configuration
 * file with tec_agent_init().
 * The second form, when used in conjunction with tec_create_handle_r(),
 * allows for multiple simultaneous configurations to exist for multiple
 * connections.
 */
EXTERN_DECL(char *tec_agent_getenv, (char *));
EXTERN_DECL(char *tec_agent_getenv_r, (tec_handle_t, char *));

/*
 * Establishes and returns a 'handle' used for several operations, including
 * sending events to an event server.
 * 
 * The second form, tec_create_handle_r(), is the same as the original first
 * form, except that the first parameter is the name of a configuration file.
 * With the second form, tec_agent_init() is automatically performed internally
 * and does not need to be invoked by the calling program.  This allows the
 * information from a configuration file to be associated with a specific
 * tec_handle_t rather than being a global variable.  This, in turn, allows
 * multiple connections to each have their own configuration (e.g., a process
 * could send events to 3 different destinations and have unique buffering,
 * filtering, and other configuration set for each one).
 */
#define tec_create_handle(_loc,_prt,_ow,_typ) tec_create_handle_r(NULL, _loc,_prt,_ow,_typ)

EXTERN_DECL(tec_handle_t tec_create_handle_r, (char *, char *, unsigned short,
                                               int, tec_delivery));

/*
 * Releases a handle created by tec_create_handle(), closes any open
 * communication channels and frees associated memory.
 */
EXTERN_DECL(void tec_destroy_handle, (tec_handle_t));

/*
 * Sends the given event string to the event server.
 */
EXTERN_DECL(long tec_put_event, (tec_handle_t, char *));

/* Using HANDLE as the platform independent name file handle/descriptor. */
#ifdef WIN32
#	include <windows.h>
#else
#	define HANDLE int
#  define INVALID_HANDLE_VALUE -1
#endif

/*
 * Returns the size of the buffer file in bytes.  The buffer file contains
 * events that could not be sent to the event server when they were passed to
 * tec_put_event().
 * If the second parameter (buffer file name) is NULL, the buffer file specified
 * in the configuration file is used.
 * If the third parameter (file handle/descriptor pointer) is not NULL, an open
 * file handle or descriptor to the buffer file is returned.  The caller must
 * close this handle.
 */
EXTERN_DECL(long tec_event_buffer_size, (tec_handle_t, char *, HANDLE *));

#endif
